/*
   termmode.c: terminal mode
*/
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include "var.h"
#include "termmode.h"

static int err_busy ( void );
static int err_bug ( void );

static int term_entry ( void );
static int term_inquiry ( char * );
static int term_inquiry_null ( char * );
static int term_help ( void );
static int term_defectlist ( int, char * );
static int term_modepage ( char * );
static int term_capacity ( char * );
static int term_unitready ( char * );

static char *devnames[] = {
  "/dev/sga", "/dev/sgb", "/dev/sgc", "/dev/sgd", "/dev/sge",
  "/dev/sgf", "/dev/sgg", "/dev/sgh", "/dev/sgi", "/dev/sgj",
  "/dev/sgk", "/dev/sgl", "/dev/sgm", "/dev/sgn", "/dev/sgo",
  "/dev/sgp", "/dev/sgq", "/dev/sgr", "/dev/sgs", "/dev/sgt",
  "/dev/sgu", "/dev/sgv", "/dev/sgw", "/dev/sgx", "/dev/sgy",
  "/dev/sgz", "" };
static int istty = 1;

#define PROMPT "scsitool> "
#define COPYRIGHT "scsitool Ver.0.5\n"\
"1996,1997 Copyright (C) by JR0BAK = Kunihiko IMAI\n"\
"                           e-mail: hab00237@niftyserve.or.jp\n\n"

/*
   term_init.h: initialize for terminal mode
*/
void term_init ()
{
  int p_ctoterm[2], p_termtoc[2];

  /*
     create a pipe pair
  */
  switch ( pipe ( p_ctoterm ) ) {
  case 0: /* normall end, nothing to do */
    break;
  case EMFILE:
  case ENFILE: err_busy(); break;
  case EFAULT:
  default: err_bug(); break;
  }

  switch ( pipe ( p_termtoc ) ) {
  case 0: /* normall end, nothing to do */
    break;
  case EMFILE:
  case ENFILE:
    err_busy(); break;
  case EFAULT:
  default:
    err_bug(); break;
  }

  /*
     disconnect/reconnect stream
  */
  istty = isatty ( 0 );
  tty_in  = fdopen ( dup ( 0 ), "r" );
  tty_out = fdopen ( dup ( 1 ), "w" );
  close ( 0 ); close ( 1 );

  if ( dup2 ( p_ctoterm[0], 0 ) != 0 ) err_busy();
  if ( dup2 ( p_termtoc[1], 1 ) != 1 ) err_busy();
  cmd_out = fdopen ( p_ctoterm[1], "w" );
  cmd_in = fdopen ( p_termtoc[0], "r" );
}



/*
   term_mode: command interpreter
*/
void term_mode ()
{
  char buf[1024];
  char *eol;
  char device[256];
  device[0] = '\0';

  if ( istty ) {
    fputs ( COPYRIGHT, tty_out );
    term_entry ();
  }

  while ( 1 ) {
    if ( istty )
      fputs ( PROMPT, tty_out );
    fflush ( tty_out );
    fgets ( buf, 1024, tty_in );
    if ( feof ( tty_in ) || ( strstr ( buf, "q" ) == buf ) ) break;
    
    eol = strchr ( buf, '\n' );
    if ( eol ) *eol = '\0';

    if ( strncmp ( buf, "c", 1 ) == 0 ) {
      term_capacity ( device );

    } else if ( strncmp ( buf, "d", 1 ) == 0) {
      if ( strchr ( buf, ' ' ) ) { /* specify device name */
	strcpy ( device, strchr ( buf, ' ' ) + 1 );
      }
      fprintf ( tty_out, "current device is '%s'\n", device );
      term_inquiry_null ( device );
      
    } else if ( strncmp ( buf, "f", 1 ) == 0 ) {
      term_format ( device );

    } else if ( strncmp ( buf, "g", 1 ) == 0 ) {
      term_defectlist ( 'g', device );

    } else if ( (strncmp ( buf, "h", 1 ) == 0) ||
	       (strncmp ( buf, "?", 1 ) == 0) ) {
      term_help ();

    } else if ( strncmp ( buf, "i", 1 ) == 0) {
      term_inquiry ( device );

    } else if ( strncmp ( buf, "m", 1 ) == 0 ) {
      term_modepage ( device );

    } else if ( strncmp ( buf, "p", 1 ) == 0 ) {
      term_defectlist ( 'p', device );

    } else if ( strncmp ( buf, "s", 1 ) == 0 ) {
      term_entry ();

    } else if ( strncmp ( buf, "t", 1 ) == 0 ) {
      term_unitready ( device );

    } else if ( buf[0] == '\0' ) {
      /* no operation */

    } else {
      fputs ( "*** can't recognize the command.\n", tty_out );
    }
  }
}



/*
   term_entry: get SCSI generic device entries
*/
int term_entry ( void )
{
  int ndev;
  char buf[1024];

  for ( ndev = 0; *devnames[ndev] != 0; ++ndev ) {
    inquiry ( devnames[ndev] );
    fgets ( buf, 1024, cmd_in );
    if ( strncmp ( buf , "***", 3 ) == 0 ) break;
    
    fputs ( devnames[ndev], tty_out );
    fputs ( ":\t", tty_out );

    while ( strncmp ( buf, "===", 3 ) ) {
      if ( strncmp ( buf, "set vender", 10 ) == 0 ) {
	fwrite ( strchr ( buf, '{' ) + 1, sizeof(char), 8, tty_out );
	fputs ( " ", tty_out );
      } else if ( strncmp ( buf, "set product", 11 ) == 0 ) {
	fwrite ( strchr ( buf, '{' ) + 1, sizeof(char), 16, tty_out );
	fputc ( '\n', tty_out );
      }
      fgets ( buf, 1024, cmd_in );
    }
  }
   return ndev;
}



/*
   term_inquiry: inquiry
*/
int term_inquiry ( char *device )
{
  char buf[1024];

  inquiry ( device );
  fgets ( buf, 1024, cmd_in );
  if ( strncmp ( buf, "***", 3 ) == 0 ) {
    fprintf ( tty_out, "*** can't inquiry of device '%s'\n", device );
    return;
  }
  while ( strncmp ( buf, "===", 3 ) ) {
    fputs ( buf+4, tty_out );
    fgets ( buf, 1024, cmd_in );
  }
}


int term_inquiry_null ( char *device )
{
  char buf[1024];

  inquiry ( device );
  fgets ( buf, 1024, cmd_in );
  if ( strncmp ( buf, "***", 3 ) == 0 ) {
    return;
  }
  while ( strncmp ( buf, "===", 3 ) ) {
    fgets ( buf, 1024, cmd_in );
  }
}

/*
   term_help: terminal mode help
*/
int term_help ( void )
{
  fputs ( "capacity            \t'Read Capacity' data\n"
	  "device <device-name>\tset target device is <device-name>\n" 
	  "format              \tphysical format the device.\n"
	  "glist               \tgrown defect list (G-LIST).\n"
	  "help                \tshow this message.\n"
	  "inquiry             \t'inquiry' data\n"
	  "show                \tshow available devices.\n"
	  "plist               \tprimary defect list (P-LIST).\n"
	  "quit                \tquit this command.\n"
	  "testunit            \t'Test Unit Ready' benchmark\n"
	 , tty_out );
}

/*
   term_format: format unit
*/
int term_format ( char *device )
{
  char buf[1024];

  fprintf ( tty_out, "device '%s' will be formatted. Are you sure (y/[N]) ? ",
	   device );
  fflush ( tty_out );
  fgets ( buf, 1024, tty_in );
  if ( strcmp ( buf, "y\n" ) ) return -1;

  fputs ( "Now formatting the unit.... Please wait patient...\n", tty_out );
  fflush ( tty_out );
  fputs ( "===\n", cmd_out ); fflush ( cmd_out );
  format_unit ( device );

  fgets ( buf, 1024, cmd_in );
  while ( strncmp ( buf, "===", 3 ) ) {
    fputs ( buf + 4, tty_out );
    fgets ( buf, 1024, cmd_in );
  }
  return 0;
}



/*
   term_defectlist: get defect list
*/
int term_defectlist ( int pg, char *device )
{
  char buf[1024];
  char *begin, *end;
  int count = 0;
  read_defect ( pg, 'p', device );

  fgets ( buf, 1024, cmd_in );
  while ( strncmp ( buf, "===", 3 ) ) {
    if ( strncmp ( buf, "***", 3 ) == 0 ) {
      fprintf ( tty_out,
	       "*** fail to get defect list of device '%s'\n", device );
      return -1;
    }
    if ( strstr ( buf, "list" ) ) {
      begin = strchr ( buf, '\"' );
      end = strchr ( begin + 1, '\"' );
      fwrite ( begin + 1, sizeof(char), end - begin - 1, tty_out );
      fputc ( '\n', tty_out );
      ++count;
    }
    fgets ( buf, 1024, cmd_in );
  }
  fprintf ( tty_out, "Total %d blocks\n", count );
}


/*
   term_modepage: mode page list
*/
int term_modepage ( char *device )
{
  char buf[1024];

  mode_page ( device );

  fgets ( buf, 1024, cmd_in );
  if ( strncmp ( buf, "***", 3 ) == 0 ) {
    fprintf ( tty_out, "*** can't get mode page list of device '%s'\n",
	     device );
    return -1;
  }
  while ( strncmp ( buf, "===", 3 ) ) {
    fputs ( buf, tty_out );
    fgets ( buf, 1024, cmd_in );
  }
}



/*
   term_capacity: show results of 'Read Capacity' command
*/
int term_capacity ( char *device )
{
  char buf[1024];

  read_capacity ( device );

  fgets ( buf, 1024, cmd_in );
  if ( strncmp ( buf, "***", 3 ) == 0 ) {
    fprintf ( tty_out, "*** can't get mode page list of device '%s'\n",
	     device );
    return -1;
  }
  while ( strncmp ( buf, "===", 3 ) ) {
    fputs ( buf+4, tty_out );
    fgets ( buf, 1024, cmd_in );
  }
}


/*
   term_unitready: show results of 'Test Unit Ready' Benchmark
*/
int term_unitready ( char *device )
{
  char buf[1024];

  unitbench ( device, 5 );

  fgets ( buf, 1024, cmd_in );
  if ( strncmp ( buf, "***", 3 ) == 0 ) {
    fprintf ( tty_out, "*** can't get mode page list of device '%s'\n",
	     device );
    return -1;
  }
  while ( strncmp ( buf, "===", 3 ) ) {
    fputs ( buf+4, tty_out );
    fgets ( buf, 1024, cmd_in );
  }
}

/*
   err_busy: error handle
*/
int err_busy ( void )
{
  fputs ( "*** System is too busy.\n", tty_out );
  exit ( -1 );
}

/*
   err_bug: error handle
*/
int err_bug( void )
{
  fputs ( "*** Fail to create pipe.\n" \
	 "*** Something is wrong.\n", tty_out );
  exit ( -1 );
}
